{-# LANGUAGE GADTs #-}
module Dependency
  ( TypedRun(..)
  , Dependency(..)
  , Token(..)
  , DependencyUntyped
  , makeUntyped
  , tupleTokens
  , zipTokens
  ) where

import Types

-- import Unsafe.Coerce (unsafeCoerce)
import Language.Haskell.TH.Syntax

-- data TypedRun a b where
--   Function :: IsFunctionId f => f -> TypedRun a b
--   FunctionIO :: IsFunctionIOId f => f -> TypedRun a b
--   Function :: IsFunction f a b => f -> TypedRun a b
--   FunctionIO :: IsFunctionIO f a b => f -> TypedRun a b
--   Inject :: Lift b => b -> TypedRun () b
--   GetListElem :: TypedRun [b] b
--   SetListElem :: TypedRun a [a]
--   MapComp :: [DependencyUntyped] -> TypedRun [a] [b]
--   FilterComp :: TypedRun ([a], [Bool]) [a]

-- deriving instance (Show a, Show b) => Show (TypedRun a b)
-- deriving instance (Lift a, Lift b) => Lift (TypedRun a b)

-- data TypedRun where
--   -- Function :: IsFunctionId f => f -> TypedRun
--   -- FunctionIO :: IsFunctionIOId f => f -> TypedRun
--   Function :: Function -> TypedRun
--   FunctionIO :: FunctionIO -> TypedRun
--   -- Inject :: (Show b, Lift b) => b -> TypedRun
--   -- InjectString :: String -> TypedRun
--   -- InjectImageConversionSettings :: ImageConversionSettings -> TypedRun
--   -- Inject :: (Show a, Lift a, Witness w a) => w -> a -> TypedRun
--   Inject :: Value -> TypedRun
--   GetListElem :: TypedRun
--   SetListElem :: TypedRun
--   MapComp :: [DependencyUntyped] -> TypedRun
--   FilterComp :: TypedRun

-- deriving instance Show TypedRun
-- deriving instance Lift TypedRun

data TypedRun = Function Function
              | FunctionIO FunctionIO
              | Inject Value
              | GetListElem
              | SetListElem
              | MapComp [DependencyUntyped]
              | FilterComp
  deriving (Show, Lift)

data Token a where
  Token :: Int -> Token a
  TupleToken :: Token a -> Token b -> Token (a, b)
  ZipToken :: Token [a] -> Token [b] -> Token [(a, b)]
  NoToken :: Token ()

deriving instance Show (Token a)
deriving instance Lift (Token a)

tupleTokens :: (Show a, Show b) => Token a -> Token b -> Token (a, b)
tupleTokens = TupleToken

zipTokens :: (Show a, Show b) => Token [a] -> Token [b] -> Token [(a, b)]
zipTokens = ZipToken

-- data Dependency a b = Dependency (Token a) (TypedRun a b) (Token b)
--   deriving (Show)

data Dependency a b = Dependency (Token a) TypedRun (Token b)
  deriving (Show)

deriving instance (Lift a, Lift b) => Lift (Dependency a b)

-- type DependencyUntyped = Dependency () ()

-- makeUntyped :: (Show a, Show b, Lift a, Lift b) => Dependency a b -> DependencyUntyped
-- makeUntyped dep = unsafeCoerce dep

data TokenUntyped = UToken Int
                  | UTupleToken TokenUntyped TokenUntyped
                  | UZipToken TokenUntyped TokenUntyped
                  | UNoToken
  deriving (Show, Lift)

makeUntypedToken :: Token a -> TokenUntyped
makeUntypedToken = \case
  Token i -> UToken i
  TupleToken a b -> UTupleToken (makeUntypedToken a) (makeUntypedToken b)
  ZipToken a b -> UZipToken (makeUntypedToken a) (makeUntypedToken b)
  NoToken -> UNoToken

-- data DependencyUntyped where
--   DependencyUntyped :: TokenUntyped -> TypedRun -> TokenUntyped -> String -> DependencyUntyped

data DependencyUntyped = DependencyUntyped TokenUntyped TypedRun TokenUntyped
  deriving (Show, Lift)

-- instance Show DependencyUntyped--  where
--   -- show (DependencyUntyped _ _ _ s) = s

-- instance Lift DependencyUntyped

makeUntyped :: (Show a, Show b, Lift a, Lift b) => Dependency a b -> DependencyUntyped
makeUntyped (Dependency ta tr tb) =
  DependencyUntyped (makeUntypedToken ta) tr (makeUntypedToken tb)
